package org.ow2.bonita.persistence.db;

import java.util.HashSet;
import java.util.Set;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.job.Timer;
import org.ow2.bonita.definition.GlobalClassData;
import org.ow2.bonita.definition.PackageClassData;
import org.ow2.bonita.definition.PackageDependency;
import org.ow2.bonita.definition.XpdlProcess;
import org.ow2.bonita.facade.uuid.PackageDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessDefinitionUUID;
import org.ow2.bonita.facade.uuid.ProcessInstanceUUID;
import org.ow2.bonita.facade.uuid.TaskUUID;
import org.ow2.bonita.persistence.RuntimeDbSession;
import org.ow2.bonita.runtime.ActivityInstanceKey;
import org.ow2.bonita.runtime.PackageVersion;
import org.ow2.bonita.runtime.ProcessInstanceKey;
import org.ow2.bonita.runtime.ProcessVersion;
import org.ow2.bonita.runtime.TaskRunTime;
import org.ow2.bonita.runtime.XpdlExecution;
import org.ow2.bonita.runtime.XpdlInstance;
import org.ow2.bonita.services.Repository;
import org.ow2.bonita.util.EnvTool;
import org.ow2.novabpm.util.Misc;

/**
 * @author Marc Blachon, Guillaume Porcher, Charles Souillard, Miguel Valdes, Pierre Vigneras
 */
public class DbRepository implements Repository {

  private final String persistenceServiceName;
  private final DbRepositoryBuffer buffer = new DbRepositoryBuffer();

  private RuntimeDbSession getRuntimeDbSession() {
    return EnvTool.getRuntimeDbSession(this.persistenceServiceName);
  }

  public DbRepository(final String persistenceServiceName) {
    this.persistenceServiceName = persistenceServiceName;
  }

  //INSTANCE REPOSITORY
  public XpdlInstance getXpdlInstance(final ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    if (this.buffer.isInstanceRemoved(instanceUUID)) {
      return null;
    }
    XpdlInstance instance = this.buffer.getInstance(instanceUUID);
    if (instance != null) {
      return instance;
    }
    instance = getRuntimeDbSession().getXpdlInstance(instanceUUID);
    if (instance != null) {
      this.buffer.addInstance(instance);
    }
    return instance;
  }

  public Set<XpdlInstance> getXpdlInstances(final ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);
    final Set<XpdlInstance> dbInstances = getRuntimeDbSession().getXpdlInstances(processUUID);
    for (final XpdlInstance instance : dbInstances) {
      if (!this.buffer.containsInstance(instance.getUUID()) && !this.buffer.isInstanceRemoved(instance.getUUID())) {
        this.buffer.addInstance(instance);
      }
    }
    final Set<XpdlInstance> result = new HashSet<XpdlInstance>();
    for (final XpdlInstance instance : this.buffer.getInstances().values()) {
      if (instance.getProcessDefinitionUUID().equals(processUUID)) {
        result.add(instance);
      }
    }
    return result;
  }

  public void storeXpdlInstance(final XpdlInstance instance) {
    Misc.checkArgsNotNull(instance);
    getRuntimeDbSession().save(instance);
    this.buffer.addInstance(instance);
  }

  public XpdlInstance removeXpdlInstance(final ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    final XpdlInstance instance = getXpdlInstance(instanceUUID);
    if (instance == null) {
      return null;
    }
    this.buffer.removeInstance(instance);
    getRuntimeDbSession().delete(instance);
    return instance;
  }

  private XpdlExecution getExecOnNode(
      final XpdlExecution exec,
      final String nodeName,
      final String iterationId,
      final String activityInstanceId) {
    Misc.checkArgsNotNull(exec, nodeName, iterationId, activityInstanceId);
    if (exec.getExecutions() == null || exec.getExecutions().isEmpty()) {
      if (exec.getNode() != null
          && exec.getNode().getName().equals(nodeName)
          && exec.getIterationId().equals(iterationId)
          && exec.getActivityInstanceId().equals(activityInstanceId)) {
        return exec;
      }
    } else {
      for (final Execution child : exec.getExecutions()) {
        final XpdlExecution found = getExecOnNode((XpdlExecution) child, nodeName, iterationId, activityInstanceId);
        if (found != null) {
          return found;
        }
      }
    }
    return null;
  }

  public XpdlExecution getExecutionPointingOnNode(
      final ProcessInstanceUUID instanceUUID, final String nodeName, final String iterationId, final String activityInstanceId) {
    Misc.checkArgsNotNull(instanceUUID, nodeName, iterationId, activityInstanceId);

    final XpdlInstance instance = this.buffer.getInstance(instanceUUID);
    if (instance != null) {
      return getExecOnNode(instance.getRootExecution(), nodeName, iterationId, activityInstanceId);
    }
    final XpdlExecution exec =
      getRuntimeDbSession().getExecutionPointingOnNode(instanceUUID, nodeName, iterationId, activityInstanceId);
    if (exec != null && !this.buffer.isInstanceRemoved(exec.getXpdlInstance().getUUID())) {
      return exec;
    }
    return null;
  }

  //PROCESS REPOSITORY
  public XpdlProcess getXpdlProcess(final ProcessDefinitionUUID processUUID) {
    Misc.checkArgsNotNull(processUUID);

    if (this.buffer.isProcessRemoved(processUUID)) {
      return null;
    }
    XpdlProcess process = this.buffer.getProcess(processUUID);
    if (process != null) {
      return process;
    }
    process = getRuntimeDbSession().getXpdlProcess(processUUID);
    if (process != null) {
      this.buffer.addProcess(process);
    }
    return process;
  }

  public void storeXpdlProcess(final XpdlProcess process) {
    Misc.checkArgsNotNull(process);
    this.buffer.addProcess(process);
    getRuntimeDbSession().save(process);
  }

  public String getLastProcessVersion(final String processId, final String packageId) {
    Misc.checkArgsNotNull(processId, packageId);
    final ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId, packageId);
    if (pv != null) {
      return pv.getProcessVersion();
    }
    return "";
  }

  public void storeProcessVersion(final String processId, final String processVersion, final String packageId) {
    Misc.checkArgsNotNull(processId, processVersion, packageId);
    ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId, packageId);
    if (pv == null) {
      pv = new ProcessVersion(processId, processVersion, packageId);
      getRuntimeDbSession().save(pv);
    } else {
      pv.setProcessVersion(processVersion);
    }
  }

  public void removeProcessVersion(final String processId, final String processVersion, final String packageId) {
    Misc.checkArgsNotNull(processId, processVersion, packageId);
    final ProcessVersion pv = getRuntimeDbSession().getLastProcessVersion(processId, packageId);
    if (pv != null) {
      final boolean stillExist = pv.removeVersion(processVersion);
      if (!stillExist) {
        getRuntimeDbSession().delete(pv);
      }
    }
  }

  public String getLastPackageVersion(final String packageId) {
    final PackageVersion pv = getRuntimeDbSession().getLastPackageVersion(packageId);
    if (pv != null) {
      return pv.getPackageVersion();
    }
    return "";
  }
  public void storePackageVersion(final String packageId, final String packageVersion) {
    PackageVersion pv = getRuntimeDbSession().getLastPackageVersion(packageId);
    if (pv == null) {
      pv = new PackageVersion(packageId, packageVersion);
      getRuntimeDbSession().save(pv);
    } else {
      pv.setPackageVersion(packageVersion);
    }
  }
  public void removePackageVersion(final String packageId, final String packageVersion) {
    final PackageVersion pv = getRuntimeDbSession().getLastPackageVersion(packageId);
    if (pv != null) {
      final boolean stillExist = pv.removeVersion(packageVersion);
      if (!stillExist) {
        getRuntimeDbSession().delete(pv);
      }
    }
  }

  public XpdlProcess removeXpdlProcess(final ProcessDefinitionUUID processUUID) {
    final XpdlProcess p = getXpdlProcess(processUUID);
    if (p == null) {
      return null;
    }
    getRuntimeDbSession().delete(p);
    return p;
  }


  //TASK REPOSITORY
  public void addTask(final TaskRunTime task) {
    Misc.checkArgsNotNull(task);
    this.buffer.addTask(task);
    getRuntimeDbSession().save(task);
  }

  public TaskRunTime getTask(final TaskUUID taskUUID) {
    Misc.checkArgsNotNull(taskUUID);

    if (this.buffer.isTaskRemoved(taskUUID)) {
      return null;
    }
    TaskRunTime task = this.buffer.getTask(taskUUID);
    if (task != null) {
      return task;
    }
    task = getRuntimeDbSession().getTask(taskUUID);
    if (task != null) {
      this.buffer.addTask(task);
    }
    return task;
  }

  public Set<TaskRunTime> getInstanceTasks(final ProcessInstanceUUID instanceUUID) {
    Misc.checkArgsNotNull(instanceUUID);
    if (this.buffer.isInstanceRemoved(instanceUUID)) {
      return new HashSet<TaskRunTime>();
    }

    final Set<TaskRunTime> dbTasks = getRuntimeDbSession().getInstanceTasks(instanceUUID);
    for (final TaskRunTime task : dbTasks) {
      if (!this.buffer.isTaskRemoved(task.getUUID())) {
        this.buffer.addTask(task);
      }
    }

    final Set<TaskRunTime> result = new HashSet<TaskRunTime>();
    for (final TaskRunTime task : this.buffer.getTasks().values()) {
      if (task.getInstanceUUID().equals(instanceUUID)) {
        result.add(task);
      }
    }

    return result;
  }

  public void removeTask(final TaskRunTime task) {
    Misc.checkArgsNotNull(task);
    this.buffer.removeTask(task);
    getRuntimeDbSession().delete(task);
  }

  //PACKAGE CLASS DATA
  public PackageClassData getPackageClassData(final PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);

    if (this.buffer.isPackageClassDataRemoved(packageUUID)) {
      return null;
    }
    PackageClassData packCD = this.buffer.getPackageClassData(packageUUID);
    if (packCD != null) {
      return packCD;
    }
    packCD = getRuntimeDbSession().getPackageClassData(packageUUID);
    if (packCD != null) {
      this.buffer.addPackageClassData(packCD);
    }
    return packCD;
  }
  public void removePackageClassData(final PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    final PackageClassData packCD = getPackageClassData(packageUUID);
    if (packCD == null) {
      return;
    }
    this.buffer.removePackageClassData(packCD);
    getRuntimeDbSession().delete(packCD);
  }
  public void storePackageClassData(final PackageClassData packageCD) {
    Misc.checkArgsNotNull(packageCD);
    getRuntimeDbSession().save(packageCD);
    this.buffer.addPackageClassData(packageCD);
  }

  //GLOBAL CLASS DATAS
  public GlobalClassData getGlobalClassData(final String className) {
    Misc.checkArgsNotNull(className);
    if (this.buffer.isGlobalClassDataRemoved(className)) {
      return null;
    }

    GlobalClassData globalCD = this.buffer.getGlobalClassData(className);
    if (globalCD != null) {
      return globalCD;
    }
    globalCD = getRuntimeDbSession().getGlobalClassData(className);
    if (globalCD != null) {
      this.buffer.addGlobalClassData(globalCD);
    }
    return globalCD;
  }

  public void removeGlobalClassData(final String className) {
    Misc.checkArgsNotNull(className);
    final GlobalClassData gcd = getGlobalClassData(className);
    if (gcd == null) {
      return;
    }
    this.buffer.removeGlobalClassData(gcd);
    getRuntimeDbSession().delete(gcd);
  }

  public void storeGlobalClassData(final GlobalClassData globalClassData) {
    Misc.checkArgsNotNull(globalClassData);
    getRuntimeDbSession().save(globalClassData);
    this.buffer.addGlobalClassData(globalClassData);
  }


  //DEPENDENT PACKAGES
  public Set<PackageDependency> getPackageDependencies(final String className) {
    Misc.checkArgsNotNull(className);
    return getRuntimeDbSession().getPackageDependencies(className);
  }
  public void addPackageDependencies(final PackageDefinitionUUID packageUUID, final Set<String> classNames) {
    Misc.checkArgsNotNull(packageUUID);
    getRuntimeDbSession().save(new PackageDependency(packageUUID, classNames));
  }
  public PackageDependency getPackageDependency(final PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    return getRuntimeDbSession().getPackageDependency(packageUUID);
  }
  public void removePackageDependencies(final PackageDefinitionUUID packageUUID) {
    Misc.checkArgsNotNull(packageUUID);
    final PackageDependency packageDependency = getPackageDependency(packageUUID);
    if (packageDependency == null) {
      return;
    }
    getRuntimeDbSession().delete(packageDependency);
  }

  public XpdlProcess findLatestProcessById(final String subProcessId) {
    return getRuntimeDbSession().findLatestProcessById(subProcessId);
  }
  public Set<Timer> getInstanceTimers(final ProcessInstanceUUID instanceUUID) {
    return getRuntimeDbSession().getInstanceTimers(instanceUUID);
  }
  public void removeTimer(final Timer timer) {
    EnvTool.getTimerSession().cancel(timer);
  }

  public synchronized long getNextActivityInstanceNb(final ProcessInstanceUUID instanceUUID,
      final String activityName) {
    final ActivityInstanceKey nextPik = new ActivityInstanceKey(instanceUUID, activityName);
    getRuntimeDbSession().save(nextPik);
    getRuntimeDbSession().delete(nextPik);
    return nextPik.getNumber();
  }

  public synchronized long getNextProcessInstanceNb(final ProcessDefinitionUUID processUUID) {
    final ProcessInstanceKey nextPik = new ProcessInstanceKey(processUUID);
    getRuntimeDbSession().save(nextPik);
    getRuntimeDbSession().delete(nextPik);
    return nextPik.getNumber();
  }

}
